default_prefix = "SHARE"

known_chains = {
    "Shareberry-Testnet": {
        "chain_id": "a3ee3a34dd82dcd8e4f9ed14202c6d35734ea4c8d9c0e9daee48ec859d135b6c",
        "prefix": "SHARE",
        "steem_symbol": "SHTO",
        "sbd_symbol": "UNDEFINED",
        "vests_symbol": "SHPO",
    },
    "Shareberry": {
        "chain_id": "a5f5e08a1f98a474b3fd943f07fa69ae5f95307c8e94e5913647f627182e2b93",
        "prefix": "SHARE",
        "steem_symbol": "SHTO",
        "sbd_symbol": "UNDEFINED",
        "vests_symbol": "SHPO",
    },
}
